import unittest
import numpy as np
import gzip
import aidge_core
import aidge_onnx
import aidge_backend_cpu
import aidge_quantization
import sys
from pathlib import Path

"""
    Unit test for the PTQ pipeline
    ==============================
    This script is designed to test and validate the accuracy of five small topologies on the MNIST dataset :
    ["MiniResNet", "ConvNet", "BranchNetV4", "TestNet", "MLP"]
    It compares the results for three configurations : baseline, quantization, and quantization with single shift. 
    The value of delta represents the tolerance of the tests.
"""
aidge_core.Log.set_console_level(aidge_core.Level.Error)  # Reduce useless logs
# --------------------------------------------------------------
# CONFIGURATION
# --------------------------------------------------------------

NB_SAMPLES   = 1000
SAMPLE_SHAPE = (1, 1, 28, 28)
NB_BITS      = 4
TARGET_TYPE  = aidge_core.dtype.int32
CLIPPING     = aidge_quantization.Clipping.MSE
NO_QUANT     = False
OPTIM_SIGNS  = True
FOLD_GRAPH   = True
DELTA        = 0.05

EXPECTED_RESULTS = {
    "MiniResNet.onnx"  : (95.4, 94.4, 95.0),
    "ConvNet.onnx"     : (97.9, 97.2, 96.7),
    "BranchNetV4.onnx" : (93.8, 92.7, 93.7),
    "TestNet.onnx"     : (95.5, 94.0, 94.5),
    "MLP.onnx"         : (94.7, 92.9, 93.8)
}

# --------------------------------------------------------------
# UTILS
# --------------------------------------------------------------

def propagate(model, scheduler, sample):
    sample = np.reshape(sample, SAMPLE_SHAPE)
    input_tensor = aidge_core.Tensor(sample)
    scheduler.forward(True, [input_tensor])
    output_node = model.get_output_nodes().pop()
    output_tensor = output_node.get_operator().get_output(0)
    return np.array(output_tensor)

def compute_accuracy(model, samples, labels):
    scheduler = aidge_core.SequentialScheduler(model)
    acc = sum(labels[i] == np.argmax(propagate(model, scheduler, x)) for i, x in enumerate(samples))
    return acc / len(samples)

# --------------------------------------------------------------
# TEST CLASS
# --------------------------------------------------------------

class TestQuantization(unittest.TestCase):

    def setUp(self):
        curr_file_dir = Path(__file__).parent.resolve()
        self.samples = np.load(gzip.GzipFile(curr_file_dir / 'assets/mnist_samples.npy.gz', "r"))
        self.labels = np.load(gzip.GzipFile(curr_file_dir / 'assets/mnist_labels.npy.gz', "r"))
        self.quant_samples = np.round(self.samples.copy() * (2**(NB_BITS-1)-1))

    def run_model_test(self, model_name):

        expected_base, expected_quant, expected_quant_ssa = EXPECTED_RESULTS[model_name]

        # load the model ...

        model_path = Path(__file__).parent / "assets" / model_name
        model = aidge_onnx.load_onnx(model_path, verbose=False)
        aidge_core.remove_flatten(model)

        model.set_datatype(aidge_core.dtype.float32)
        model.set_backend("cpu")
        
        # create the tensor subset

        tensors = [aidge_core.Tensor(np.reshape(sample, SAMPLE_SHAPE)) for sample in self.samples[:NB_SAMPLES]]

        # BASELINE ACCURACY

        base_accuracy = compute_accuracy(model, tensors, self.labels[:NB_SAMPLES])
        self.assertAlmostEqual(base_accuracy * 100, expected_base, delta=DELTA, msg=f"[X] Baseline accuracy mismatch for {model_name}. Expected accuracy was: {expected_base}, but got: {base_accuracy * 100}")

        # QUANTIZED ACCURACY

        aidge_quantization.quantize_network(
            network=model,
            nb_bits=NB_BITS,
            calibration_set=tensors,
            target_type=TARGET_TYPE,
            clipping_mode=CLIPPING,
            no_quant=NO_QUANT,
            optimize_signs=OPTIM_SIGNS,
            single_shift=False,
            use_cuda=False,
            fold_graph=FOLD_GRAPH)

        quant_accuracy = compute_accuracy(model, self.quant_samples[:NB_SAMPLES], self.labels)
        self.assertAlmostEqual(quant_accuracy * 100, expected_quant, delta=DELTA, msg=f"[X] Quantized accuracy mismatch for {model_name}. Expected accuracy was: {expected_quant}, but got: {quant_accuracy * 100}")
        
        # QUANTIZED ACCURACY WITH SSA

        model = aidge_onnx.load_onnx(model_path, verbose=False)
        model.set_datatype(aidge_core.dtype.float32)
        model.set_backend("cpu")
        
        aidge_quantization.quantize_network(
            network=model,
            nb_bits=NB_BITS,
            calibration_set=tensors,
            target_type=TARGET_TYPE,
            clipping_mode=CLIPPING,
            no_quant=NO_QUANT,
            optimize_signs=OPTIM_SIGNS,
            single_shift=True,
            use_cuda=False,
            fold_graph=FOLD_GRAPH)

        quant_accuracy_ssa = compute_accuracy(model, self.quant_samples[:NB_SAMPLES], self.labels)
        self.assertAlmostEqual(quant_accuracy_ssa * 100, expected_quant_ssa, delta=DELTA, msg=f"[X] Quantized accuracy (with SSA) mismatch for {model_name}. Expected accuracy was: {expected_quant_ssa}, but got: {quant_accuracy_ssa * 100}")

    def test_models(self):
        for model in EXPECTED_RESULTS.keys():
            with self.subTest(model=model):
                self.run_model_test(model)

if __name__ == '__main__':
    unittest.main()
