/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/
#ifndef AIDGE_QUANTIZATION_QUANTIZATION_PTQ_PTQMETAOPS_H_
#define AIDGE_QUANTIZATION_QUANTIZATION_PTQ_PTQMETAOPS_H_

#include <memory>
#include <string>

#include "aidge/graph/GraphView.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/data/Data.hpp"

namespace Aidge {

    /**
     * @brief Create a Quantizer node that initially consists of a multiplier and a scaling factor.
     * @param scalingFactor The value of the multiplicative coefficient.
     * @param name Name of the Quantizer.
     */
    std::shared_ptr<Aidge::Node> Quantizer(double scalingFactor, const std::string& name);

    /**
     * @brief Given a Quantizer, multiply it's internal multiplicative coefficient by a value.
     * @param coeff The value of the multiplicative coefficient.
     */
    void multiplyScalingFactor(std::shared_ptr<Aidge::Node> quantizer, double coeff);

    /**
     * @brief Given a Quantizer, create a copy of it that has a Round node and a Clip node added
     * at its endpoint, and replace the given Quantizer by it (a swap is also done by reference). 
     * @param quantizer The quantizer to modify and replace.
     * @param clipMin the min value of the clip node.
     * @param clipMax the max value of the clip node.
     */
    void appendRoundClip(std::shared_ptr<Node>& quantizer, double clipMin, double clipMax);
    
    /**
     * @brief Given a Quantizer, create a copy of it that has the Round node removed,
     * and replace the given Quantizer by it (a swap is also done by reference). 
     * @param quantizer The quantizer to modify and replace.
     */
    void removeRound(std::shared_ptr<Node>& quantizer);

    /**
     * @brief Given a Quantizer, create a copy of it where the Mul node is replaced by
     * a Bit-Shift node, and replace the given Quantizer by it (a swap is also done by reference). 
     * @param quantizer The quantizer to modify and replace.
     */
    void replaceScalingWithBitShift(std::shared_ptr<Node>& quantizer);

    /**
     * @brief Given a Quantizer, create a copy of it that has Cast nodes inserted at it's IOs,
     * and replace the given Quantizer by it (a swap is also done by reference). The input cast 
     * node convert the input data to the internal type, while the output cast convert it back
     * to the external type.
     * @param quantizer The quantizer to modify and replace.
     * @param externalType The external data type used for the casts.
     */
    void castQuantizerIOs(std::shared_ptr<Node>& quantizer, Aidge::DataType externalType);

    /**
     * @brief Given a Quantizer, retreive the coefficient of it's Mul node.
     * @param quantizer The quantizer containing the multiplicative coefficient.
     */
    double getScalingFactor(std::shared_ptr<Aidge::Node> quantizer);

    /**
     * @brief Given a Quantizer containing a Clip node, replace its clipping values.
     * @param quantizer The quantizer containing the Clip node.
     * @param min The min clipping value.
     * @param max The max clipping value.
     */
    void setClipRange(std::shared_ptr<Aidge::Node> quantizer, double min, double max);

}

#endif /* AIDGE_QUANTIZATION_QUANTIZATION_PTQ_PTQMETAOPS_H_ */
