/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_QUANTIZATION_QUANTIZATION_PTQ_PTQ_H_
#define AIDGE_QUANTIZATION_QUANTIZATION_PTQ_PTQ_H_

#include <cstdint>  // std::uint8_t
#include <unordered_map>
#include <memory>
#include <set>
#include <string>
#include <utility>  // std::pair
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/quantization/PTQ/Clipping.hpp"
#include "aidge/graph/GraphView.hpp"

namespace Aidge {

    /**
     * @brief Set of the types of the nodes which contain affine transforms (that is Y = A.X + B)
     */
    static const std::set<std::string> affineNodeTypes({"FC", "Conv2D", "ConvDepthWise2D", "PaddedConv2D", "PaddedConvDepthWise2D"});

    /**
     * @brief Set of the types of the nodes which does not affect the PTQ process
     */
    static const std::set<std::string> seamlessNodeTypes({"LeakyReLU", "Pad2D", "MaxPooling2D", "AvgPooling2D", "PaddedMaxPooling2D", "PaddedAvgPooling2D", "GlobalAveragePooling", "Reshape", "Transpose", "Gather", "Resize"});

    /**
     * @brief Set of the types of the nodes that merge multiple branches into one
     */
    static const std::set<std::string> mergingNodeTypes({"Add", "Concat", "Sub"});

    /**
     * @brief Set of the types of the nodes that won't be quanized
     */
    static const std::set<std::string> notQuantizedNodeTypes({"Sigmoid", "Tanh"});

    /**
     * @brief Determine if a node contains an affine transform (that is Y = A.X + B)
     * @param node The node to be checked
     * @return True if the node is affine, else false.
     */
    bool isAffine(std::shared_ptr<Node> node);

    /**
     * @brief Determine if a node contains an operator that does not affect the PTQ process
     * @param node The node to be checked
     * @return True if the node is seamless, else false.
     */
    bool isSeamless(std::shared_ptr<Node> node);

    /**
     * @brief Determine if a node contains an operator that merges multiple branches into one
     * @param node The node to be checked
     * @return True if the node is merging, else false.
     */
    bool isMerging(std::shared_ptr<Node> node);

    /**
     * @brief Determine if a node contains an operator that won't be quantized
     * @param node The node to be checked
     * @return True if the node is not quantized, else false.
     */
    bool isNotQuantized(std::shared_ptr<Node> node);

    /**
     * @brief Compute the absolute max of a tensor
     * @param tensor The Tensor to process
     */
    double getTensorAbsoluteMax(std::shared_ptr<Tensor> tensor);

    /**
     * @brief Retrieve the scheduled vector of node of a graphView, without the Producer nodes.
     * @param graphView The graphView containing the nodes
     * @param verbose Whether to print the node vector or not
     * @return The scheduled vector of nodes
     */
    std::vector<std::shared_ptr<Node>> retrieveNodeVector(std::shared_ptr<GraphView> graphView, bool newSchedule = true, bool verbose = false);
    
    /**
     * @brief Inserts a scaling node below the given producer node in the graphView. 
     * @param node A shared pointer to the producer node where the scaling node will be inserted (below).
     * @param graphView A shared pointer to the graph view in which the nodes are located.
     * @return True if the scaling node was successfully inserted or the scaling factor was accumulated; False otherwise.
     */
    void insertScalingBelowProducer(std::shared_ptr<Node> node, std::shared_ptr<GraphView> graphView);

    /**
     * @brief Determine whether an input GraphView can be quantized or not.
     * @param graphView The GraphView to be checked.
     * @return True if the GraphView can be quantized, else false.
     */
    bool checkArchitecture(std::shared_ptr<GraphView> graphView);

    /**
     * @brief This function multiplies the existing scaling factor by a given coefficient. It verifies that the node is of the correct type ("Mul") 
     * and has the `isScaling` attribute. If these conditions are not met, a warning is logged.
     * @param node A shared pointer to an `Aidge::Node` object representing the node to modify.
     * @param coeff  A double representing the multiplication coefficient to apply to the scaling factor.
     */
    void multiplyScalingFactor(std::shared_ptr<Aidge::Node> node, double coeff);


    /**
     * @brief Prepare a network before the quantization is applied to it, by removing, replacing
     * or fusing the nodes that are not supported by the PTQ pipeline. 
     * @param graphView The network to prepare for the quantization
     */
    void prepareNetwork(std::shared_ptr<GraphView> graphView);

    /**
     * @brief Insert a scaling node after each affine node of the GraphView.
     * Also insert a scaling node in every purely residual branches.
     * @param graphView The GraphView containing the affine nodes.
     */
    void insertScalingNodes(std::shared_ptr<GraphView> graphView);

    /**
     * @brief Normalize the parameters of each parametrized node, so that they fit in the [-1:1] range.
     * @param graphView The GraphView containing the parametrized nodes.
     */
    void normalizeParameters(std::shared_ptr<GraphView> graphView);

    /**
     * @brief Compute the activation ranges of every affine node, given an input dataset.
     * @param graphView The GraphView containing the affine nodes, on which the inferences are performed.
     * @param calibrationSet The calibration dataset, consisting of a vector of input samples.
     * @param scalingNodesOnly Whether to restrain the retreival of the ranges to scaling nodes only or not.
     * @return A map associating each affine node name to it's corresponding output range.
     */
    std::unordered_map<std::shared_ptr<Node>, double> computeRanges(std::shared_ptr<GraphView> graphView, std::vector<std::shared_ptr<Tensor>> calibrationSet, bool scalingNodesOnly, bool useCuda);

    /**
     * @brief Normalize the activations of each affine node so that they fit in the [-1:1] range.
     * This is done by reconfiguring the scaling nodes, as well as rescaling the weights and biases tensors.
     * @param graphView The GraphView containing the affine nodes.
     * @param valueRanges The node output value ranges computed over the calibration dataset.
     */
    void normalizeActivations(std::shared_ptr<GraphView> graphView, std::unordered_map<std::shared_ptr<Node>, double> valueRanges);

    /**
     * @brief For each node, compute the sign of its input and output values.
     * The goal of the routine is to maximize the number of unsigned IOs in order to double the value resolution when possible.
     * @param graphView The GraphView to analyze.
     * @param verbose Whether to print the sign map or not.
     * @return A map associating a pair of sign to each node of the GraphView (a sign for the input and one for the output).
     */
    std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> computeSignMap(std::shared_ptr<GraphView> graphView, bool verbose);

    /**
     * @brief Quantize an already normalized (in term of parameters and activations) network.
     * @param graphView The GraphView to be quantized.
     * @param nbBits The desired number of bits of the quantization.
     * @param noQuant Whether to apply the rounding operations or not.
     * @param optimizeSigns Whether to take account of the IO signs of the operators or not.
     * @param verbose Whether to print the sign map or not.
     */
    void quantizeNormalizedNetwork(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, bool noQuant, bool optimizeSigns, bool verbose);

    /**
     * @brief Take a quantized GraphView represented in floating precision and cast it to the desired target precision. 
     * If single-shift option is set to True, the scaling nodes contained in the activation quantizers are replaced with bit-shifts.
     * @param graphView The GraphView to modify.
     * @param targetType The desired precision of the cast.
     * @param singleShift If set to True, replace the scaling-factors by bit-shifts.
     * @param bitShiftRounding If singleShift is True, specifies the kind of bit-shift roundinqg..
     * 
     */  
    void castQuantizedNetwork(std::shared_ptr<GraphView> graphView, Aidge::DataType targetType, bool singleShift /*, bool bitShiftRounding*/);

    /**
     * @brief Main quantization routine. Performs every step of the quantization pipeline.
     * @param graphView The GraphView to be quantized.
     * @param nbBits The desired number of bits of the quantization.
     * @param calibrationSet The calibration dataset used for the activations calibration.
     * @param targetType The desired data-type of the outputed GraphView.
     * @param clippingMode: Type of the clipping optimization. Can be either 'MAX', 'MSE', 'AA' or 'KL'.
     * @param noQuant Whether to apply the rounding operations or not.
     * @param optimizeSigns Whether to take account of the IO signs of the operators or not.
     * @param singleShift Whether to convert the scaling factors into powers of two. If true the approximations are compensated using the previous nodes parameters.
     * @param useCuda Whether to use the CUDA backend for performing the activation calibration or not.
     * @param foldGraph Whether to fold the parameter quantizers after the quantization or not.
     * @param verbose Whether to print internal informations about the quantization process or not.
     */
    void quantizeNetwork(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, std::vector<std::shared_ptr<Tensor>> calibrationSet, DataType targetType, Clipping clippingMode, bool noQuant, bool optimizeSigns, bool singleShift, bool useCuda, bool foldGraph, bool verbose);


    /**
     * @brief Compute the weight ranges of every affine node. Provided for debugging purposes.
     * @param graphView The GraphView containing the affine nodes.
     * @return A map associating each affine node name to it's corresponding weight range.
     */
    std::unordered_map<std::string, double> getWeightRanges(std::shared_ptr<GraphView> graphView);

    /**
     * @brief Clear the affine nodes biases. Provided form debugging purposes.
     * @param graphView The GraphView containing the affine nodes.
     */
    void clearBiases(std::shared_ptr<GraphView> graphView);

    /**
     * @brief Developement and test routine.
     * @param graphView The GraphView under test.
     */
    void devPTQ(std::shared_ptr<GraphView> graphView);
}

#endif /* AIDGE_QUANTIZATION_QUANTIZATION_PTQ_PTQ_H_ */

