/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/quantization/PTQ/CLE.hpp"

#include <cmath>    // std::abs, std::fabs, std::sqrt
#include <cstddef>  // std::size_t
#include <memory>
#include <vector>

#include "aidge/quantization/PTQ/Clipping.hpp"
#include "aidge/quantization/PTQ/PTQ.hpp"  // retrieveNodeVector

#include "aidge/graph/GraphView.hpp"
#include "aidge/scheduler/SequentialScheduler.hpp"
#include "aidge/scheduler/Scheduler.hpp"
#include "aidge/utils/Log.hpp"
#include "aidge/operator/OperatorTensor.hpp"

#include "aidge/operator/Mul.hpp"
#include "aidge/operator/ArgMax.hpp"
#include "aidge/operator/Abs.hpp"
#include "aidge/operator/Reshape.hpp"
#include "aidge/operator/Round.hpp"
#include "aidge/operator/MetaOperator.hpp"

namespace Aidge
{

static std::shared_ptr<Tensor> getWeightTensor(std::shared_ptr<Node> node)
{
    return std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(1);
}

static std::shared_ptr<Tensor> getBiasTensor(std::shared_ptr<Node> node)
{
    return std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(2);
}

static bool nodeHasBias(std::shared_ptr<Node> node)
{
    if (node->getParents().size() == 3) {
        std::shared_ptr<Tensor> biasTensor = getBiasTensor(node);
        if (biasTensor)
            return true;
    }
    return false;
}

std::shared_ptr<Aidge::Tensor> getScaledWeightTensor(std::shared_ptr<Node> node) 
{
    if (node->getParent(1)->attributes()->hasAttr("quantization.ptq.isProducerQuantizer"))
    {
        auto quantizer = node->getParent(1);

        // perform an inference on the branch

        auto graphView = Sequential({quantizer});
        graphView->add(quantizer->getParent(0));
        SequentialScheduler scheduler(graphView);
        scheduler.forward(true, {});

        // gather and return the result

        auto op = std::static_pointer_cast<MetaOperator_Op>(quantizer->getOperator());
        auto result = op->getOutput(0);
        return result;
    } 
    else 
    {
        auto result = getWeightTensor(node);        
        return result;
    }
}

void crossLayerEqualization(std::shared_ptr<GraphView> graphView, double targetDelta)
{
    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    // Check if the CLE can be applied ...

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (node->getChildren().size() > 1) {
            Log::warn(" Network have multiple branches, skipping the CLE ... ");
            return;
        }
        if (isNotQuantized(node)) {
            Log::warn(" Network contains non linear nodes, skipping the CLE ... ");
            return;
        }
    }

    Log::notice(" Applying the Cross-Layer Equalization ... ");

    // Get the vector of affine nodes

    std::vector<std::shared_ptr<Node>> affineNodeVector;
    for (std::shared_ptr<Node> node : nodeVector)
        if (isAffine(node))
            affineNodeVector.push_back(node);

    double maxRangeDelta;
    do
    {
        maxRangeDelta = 0.0;
        
        for (size_t i = 0; i < (affineNodeVector.size() - 1); i++)
        {
            // Log::notice(" node index : {} ", i);

            std::shared_ptr<Node> n1 = affineNodeVector[i];
            std::shared_ptr<Node> n2 = affineNodeVector[i+1];

            std::shared_ptr<Aidge::Tensor> w1 = getScaledWeightTensor(n1);
            std::shared_ptr<Aidge::Tensor> w2 = getScaledWeightTensor(n2);
            
            //Log::notice(" TENSOR : \n {}", *w1);

            double r1 = getTensorAbsoluteMax(w1);
            double r2 = getTensorAbsoluteMax(w2);

            double s1 = std::sqrt(r1 * r2) / r1;
            double s2 = std::sqrt(r1 * r2) / r2;

            multiplyScalingFactor(n1->getParent(1), s1);

            if (nodeHasBias(n1))
                multiplyScalingFactor(n1->getParent(2), s1);

            multiplyScalingFactor(n2->getParent(1), s2);

            double rangeDelta = std::abs(r1 - r2);
            if (rangeDelta > maxRangeDelta)
                maxRangeDelta = rangeDelta;
        }

        // Log::notice(" CLE delta = {} ", maxRangeDelta);
    }
    while (maxRangeDelta > targetDelta);
}

}