/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

 #include "aidge/quantization/PTQ/CLE.hpp"
 #include "aidge/quantization/PTQ/Clipping.hpp"
 #include "aidge/quantization/PTQ/PTQ.hpp"
 #include "aidge/operator/PTQMetaOps.hpp"
 
 #include "aidge/data/Tensor.hpp"
 #include "aidge/graph/GraphView.hpp"
 #include "aidge/graph/Node.hpp"
 #include "aidge/scheduler/SequentialScheduler.hpp"
 #include "aidge/scheduler/Scheduler.hpp"
 #include "aidge/utils/Log.hpp"
 
 #include "aidge/operator/Producer.hpp"
 #include "aidge/operator/Mul.hpp"
 #include "aidge/operator/Round.hpp"
 #include "aidge/operator/ReLU.hpp"
 #include "aidge/operator/BatchNorm.hpp"
 #include "aidge/operator/Conv.hpp"
 #include "aidge/operator/ArgMax.hpp"
 #include "aidge/operator/Reshape.hpp"
 #include "aidge/operator/MatMul.hpp"
 
 #include "aidge/recipes/Recipes.hpp"
 #include "aidge/recipes/QuantRecipes.hpp"
 #include "aidge/operator/MetaOperator.hpp"

namespace Aidge
{

static bool hasAttr(std::shared_ptr<Aidge::Node> node, std::string attr)
{
    return node->attributes()->hasAttr("quantization.ptq." + attr);
}

static void addAttr(std::shared_ptr<Aidge::Node> node, std::string attr, double value = 0.0)
{
    node->attributes()->addAttr("quantization.ptq." + attr, value);
}

bool isAffine(std::shared_ptr<Node> node)
{   
    if (affineNodeTypes.find(node->type()) != affineNodeTypes.end())
        return true;

    if ((node->type() == "MatMul") && hasAttr(node, "isWeighted"))
        return true;

    return false;
}

bool isSeamless(std::shared_ptr<Node> node)
{
    return (seamlessNodeTypes.find(node->type()) != seamlessNodeTypes.end());
}

bool isMerging(std::shared_ptr<Node> node)
{
    if (mergingNodeTypes.find(node->type()) != mergingNodeTypes.end())
        return true;

    if ((node->type() == "MatMul") && !hasAttr(node, "isWeighted"))
        return true;

    return false;
}

bool isNotQuantized(std::shared_ptr<Node> node)
{
    return (notQuantizedNodeTypes.find(node->type()) != notQuantizedNodeTypes.end());
}

void clearGraphViewInputNodes(std::shared_ptr<GraphView> graphView,DataType targetType)
{
    for (std::shared_ptr<Aidge::Node> node : graphView->inputNodes())
        for (Aidge::IOIndex_t i = node->getFirstFreeDataInput(); i < node->getNbFreeDataInputs(); i++)
            node->getOperator()->resetInput(i);
}

bool checkArchitecture(std::shared_ptr<GraphView> graphView)
{
    std::set<std::string> removedNodeTypes({"Flatten", "Softmax", "BatchNorm2D"});

    std::set<std::string> specialNodeTypes({"MatMul", "ReLU", "Producer"});

    std::set<std::string> notQuantizedNodesTypes;

    for (std::shared_ptr<Node> node : graphView->getNodes())
    {
        bool isRemoved = removedNodeTypes.find(node->type()) != removedNodeTypes.end(); 
        bool isSpecial = specialNodeTypes.find(node->type()) != specialNodeTypes.end();
        if (!isRemoved && !isSpecial && !isAffine(node) && !isSeamless(node) && !isMerging(node) && !isNotQuantized(node)) {
            Log::warn(" GraphView can't be quantized : node type {} is not supported !", node->type());
            return false;
        }

        if (isNotQuantized(node))
            notQuantizedNodesTypes.insert(node->type());
    }

    if (!notQuantizedNodesTypes.empty()) {
        std::string tokens;
        for (std::string s : notQuantizedNodesTypes)
            tokens += (s + " ");
        Log::warn(" Network contains non-linear nodes that won't be quantized : {}", tokens);
    }

    return true;
}

void prepareNetwork(std::shared_ptr<GraphView> graphView)
{
    // remove the flatten nodes

    removeFlatten(graphView);

    // handle the MatMuls

    reorderMatMulInputs(graphView);
    // matMulToFC(graphView); // not working properly atm !

    // tag the weighted nodes

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        bool isWeighted = isAffine(node);
        if (node->type() == "MatMul") 
        {
            std::shared_ptr<Node> parent = node->getParent(1);
            if (parent) 
                if (parent->type() == "Producer") 
                    isWeighted = true;
        }

        if (isWeighted)
            addAttr(node, "isWeighted");
    }

    // fuse the batchnorms

    bool containsBatchNorm = false;
    for (std::shared_ptr<Node> node : nodeVector) {
        if (node->type() == "BatchNorm") {
            containsBatchNorm = true;
            break;
        }
    }

    if (containsBatchNorm)
        fuseBatchNorm(graphView);   

    // pop the softmax

    popSoftMax(graphView);
}

static std::shared_ptr<Aidge::Node> getUniqueChild(std::shared_ptr<Aidge::Node> node) 
{
    std::set<std::shared_ptr<Aidge::Node>> childrenSet = node->getChildren();
    AIDGE_ASSERT(childrenSet.size() == 1, " Attempted to access to a unique child while the parent have multiple ones ! ");
    return *(childrenSet.begin());
}

static std::string determineBackend(std::shared_ptr<Aidge::Node> node)
{
    std::string backend = node->getOperator()->backend();

    if (backend != "")
        return backend;
    else 
    {
        // gather the parent backends

        std::set<std::string> parentBackends;
        for (auto parent : node->getParents())
            parentBackends.insert(determineBackend(parent)); // it always answers a non empty value !

        // check if we have two or more different backends gathered

        if (parentBackends.size() > 1) 
        {
            Log::warn(" Unable to determine backend of node {} due to conflicting parent ones !", node->name());
            return (*parentBackends.begin());
        }
        
        // if all parents have the same backend return it

        if (parentBackends.size() == 1) 
            return (*parentBackends.begin());
    }

    return "cpu"; // escape path when no parents are found
}

static int getInputIndex(std::shared_ptr<Node> node, std::shared_ptr<Node> parentNode)
{
    int index = 0;
    while (node->getParent(index) != parentNode) 
        index++;
    return index;
}

static void insertChildren(std::shared_ptr<Node> parent, std::shared_ptr<Node> newNode, std::shared_ptr<GraphView> graphView) 
{
    // Checking the parents always have at least 1 children

    AIDGE_ASSERT(parent->getChildren().size() > 0, " Parent node must have at least one child to insert a new node ! ");

    // Retreive children connection indexes

    std::vector<std::shared_ptr<Node>> nextNodes = parent->getChildren(0);
    std::vector<int> inputIndices(nextNodes.size());
    for (std::size_t i = 0; i < nextNodes.size(); i++) {
        inputIndices[i] = getInputIndex(nextNodes[i], parent);
    }

    // Disconnect childs from parent

    for (std::shared_ptr<Node> nextNode : nextNodes) {
        parent->removeChild(nextNode, 0);
    }

    // Insert the new node between the child and the parent

    parent->addChild(newNode, 0, 0);
    for (std::size_t i = 0; i < nextNodes.size(); i++) {
        newNode->addChild(nextNodes[i], 0, inputIndices[i]);
    }

    graphView->add(newNode);
}

void foldProducerQuantizers(std::shared_ptr<GraphView> graphView)
{
    std::vector<std::shared_ptr<Node>> producerQuantizers;
    for (std::shared_ptr<Node> node : graphView->getNodes())
        if (hasAttr(node, "isProducerQuantizer"))
            producerQuantizers.push_back(node);

    for (std::shared_ptr<Node> quantizer : producerQuantizers)
    {
        // Set the param producer to be constant

        auto paramProducer = quantizer->getParent(0);
        auto paramProducerOp = std::static_pointer_cast<Producer_Op>(paramProducer->getOperator());
        paramProducerOp->constant() = true;

        // Set the internal producers of the quantizer to be constant

        auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    
        auto microGraph = quantizerOp->getMicroGraph();
        
        for (auto producer : microGraph->getNodes())
            if (producer->type() == "Producer")
            {
                auto producerOp = std::static_pointer_cast<Producer_Op>(producer->getOperator());
                producerOp->constant() = true;  
            }   

        expandMetaOp(quantizer); // mandatory for now !!!
    }

    constantFolding(graphView);
}

void castQuantizedNetwork(std::shared_ptr<GraphView> graphView, Aidge::DataType targetType, bool singleShift /*, bool bitShiftRounding*/)
{
    std::set<Aidge::DataType> supportedFloatTypes = {DataType::Float16, DataType::Float32, DataType::Float64};
    std::set<Aidge::DataType> supportedIntTypes = {DataType::Int16, DataType::Int32, DataType::Int64};

    bool castToFloat = (supportedFloatTypes.find(targetType) != supportedFloatTypes.end());
    bool castToInt = (supportedIntTypes.find(targetType) != supportedIntTypes.end());

    if (castToFloat)
    {
        graphView->setDataType(targetType);
    }
    else if (castToInt)
    {
        if (singleShift)
        {
            Log::notice(" Replacing scaling nodes with bit-shifts ...");

            // Replace the scaling nodes with bit-shifts (activations only)

            std::set<std::shared_ptr<Node>> nodes = graphView->getNodes(); // must be called again because of removeRound() !
            for (std::shared_ptr<Node> node : nodes)
            {
                if (node->type() == "Quantizer")
                {
                    if (hasAttr(node, "isActivationQuantizer"))
                    {
                        removeRound(node);
                        replaceScalingWithBitShift(node);
                    }
                    else if (hasAttr(node, "isProducerQuantizer"))
                        castQuantizerIOs(node, targetType);
                }
            }

            // Cast the nodes (excepted the producers and quantizers) to integer precision
            nodes = graphView->getNodes();
            for (std::shared_ptr<Node> node : nodes)
                if (node->type() != "Producer" && !hasAttr(node, "isProducerQuantizer")) // TODO : double check this !
                    node->getOperator()->setDataType(targetType);
        }
        else
        {
            // Set the nodes (excepted the producers and quantizers) to have integer IOs

            std::set<std::shared_ptr<Node>> nodes = graphView->getNodes();
            for (std::shared_ptr<Node> node : nodes)
                if (node->type() != "Quantizer" && node->type() != "Producer")
                    node->getOperator()->setDataType(targetType);

            // Cast the quantizers input and outputs by inserting Cast nodes

            for (std::shared_ptr<Node> node : nodes)
                if (node->type() ==  "Quantizer")
                    castQuantizerIOs(node, targetType);
        }
    }
    else
    {
        Log::error(" Cannot cast the quantized network : target type '{}' is not supported ! ", targetType);
    }
}

double getTensorAbsoluteMax(std::shared_ptr<Tensor> tensor)
{
    std::shared_ptr<Tensor> fallback;

    // get the abs tensor

    std::shared_ptr<Tensor> absTensor = std::make_shared<Tensor>(tensor->abs());

    // flatten the abs tensor

    std::int64_t nbElement = tensor->size();

    auto reshapeOp = Reshape_Op({nbElement});
    reshapeOp.setDataType(tensor->dataType());
    reshapeOp.setBackend(tensor->backend());

    reshapeOp.associateInput(0, absTensor);
    reshapeOp.forward();
    std::shared_ptr<Tensor> flatTensor = reshapeOp.getOutput(0);
    const Tensor& localFlatTensor = flatTensor->refCastFrom(fallback, DataType::Float64, "cpu"); 

    // Get the argmax

    auto argmaxOp = ArgMax_Op(0, true, false);
    argmaxOp.setDataType(tensor->dataType());
    argmaxOp.setBackend(tensor->backend());

    argmaxOp.associateInput(0, flatTensor);
    argmaxOp.forward();
    const Tensor& argMaxTensor = argmaxOp.getOutput(0)->refCastFrom(fallback, DataType::Float64, "cpu"); 

    // Return the max

    int maxIndex = std::round(argMaxTensor.get<double>(0));

    return localFlatTensor.get<double>(maxIndex);
}

static void fixScheduling(std::vector<std::shared_ptr<Node>>& nodeVector) {

    std::vector<std::shared_ptr<Node>> correctedVector;

    for (int i = (nodeVector.size() - 1); i >= 0; --i)
    {
        std::shared_ptr<Node> node = nodeVector[i];
        bool isAlreadyInside = (std::find(correctedVector.begin(), correctedVector.end(), node) != correctedVector.end());
        if (!isAlreadyInside)
            correctedVector.push_back(node);
    }

    std::reverse(correctedVector.begin(), correctedVector.end());

    nodeVector = correctedVector;
}

static std::shared_ptr<Tensor> getWeightTensor(std::shared_ptr<Node> node)
{
    std::shared_ptr<Node> producer = node->getParent(1);

    if (producer->type() == "Quantizer") 
        producer = producer->getParent(0);

    return std::static_pointer_cast<OperatorTensor>(producer->getOperator())->getOutput(0);
}

static std::shared_ptr<Tensor> getBiasTensor(std::shared_ptr<Node> node)
{
    std::shared_ptr<Node> producer = node->getParent(2);

    if (producer->type() == "Quantizer") 
        producer = producer->getParent(0);

    return std::static_pointer_cast<OperatorTensor>(producer->getOperator())->getOutput(0);
}

std::vector<std::shared_ptr<Node>> retrieveNodeVector(std::shared_ptr<GraphView> graphView, bool newSchedule, bool verbose)
{
    std::vector<std::shared_ptr<Node>> nodeVector = graphView->getOrderedNodes();
   
    // Remove duplicate nodes. XXX Is it still needed ???

    fixScheduling(nodeVector); 

    // Remove Producers and their Scalings

    std::vector<std::shared_ptr<Node>> remainingNodes;
    for (std::shared_ptr<Node> node : nodeVector)
        if ((node->type() != "Producer") && !hasAttr(node, "isProducerQuantizer"))
            remainingNodes.push_back(node);

    nodeVector = remainingNodes;

    // Verbose

    if (verbose) 
    {
        Log::info(" NB OF NODES = {}", nodeVector.size());
        for (std::shared_ptr<Node> node : nodeVector)
            Log::info("{} {}", node->type(), node->name());
    }

    return nodeVector;    
}

static std::shared_ptr<Node> getFirstNode(std::shared_ptr<GraphView> graphView)
{
    return retrieveNodeVector(graphView)[0];
}

// TODO : enhance this by modifying OperatorImpl in "core" ...
static DataType getDataType(std::shared_ptr<Node> node)
{
    auto op = std::static_pointer_cast<OperatorTensor>(node->getOperator());
    return op->getOutput(0)->dataType();
}

// XXX double check this !
static bool nodeHasBias(std::shared_ptr<Node> node)
{
    if (node->getParents().size() == 3) {
        std::shared_ptr<Tensor> biasTensor = getBiasTensor(node);
        if (biasTensor)
            return true;
    }
    return false;
}

// TODO: rework this !
static std::shared_ptr<Node> getPreviousScalingNode(std::shared_ptr<Node> node)
{
    std::shared_ptr<Node> currNode = node;
    while(!hasAttr(currNode, "isActivationQuantizer")) {
        if (currNode->getParents().size() == 0) {
            Log::warn(" Warning : No previous Scaling node were found ! ");
            break;
        }
        currNode = currNode->getParents()[0];
    }
    return currNode;
}

void insertScalingBelowProducer(std::shared_ptr<Node> producerNode, std::shared_ptr<GraphView> graphView)
{
    std::string scalingNodeName = makeUniqueName(producerNode->name() + "_ProducerScaling", graphView);
    std::shared_ptr<Node> scalingNode = Quantizer(1.0, scalingNodeName);;
    addAttr(scalingNode, "isProducerQuantizer");

    scalingNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
    scalingNode->getOperator()->setBackend(determineBackend(producerNode)); // XXX use the producer parent instead ???

    insertChildren(producerNode, scalingNode, graphView);
}

void insertProducerScalingNodes(std::shared_ptr<GraphView> graphView)
{
    std::set<std::shared_ptr<Node>> nodeSet = graphView->getNodes();

    for (std::shared_ptr<Node> node : nodeSet)
    {
        if (isAffine(node))
        {
            insertScalingBelowProducer(node->getParent(1), graphView);
            if (nodeHasBias(node))
                insertScalingBelowProducer(node->getParent(2), graphView);
        }
    }
}

// XXX HERE : Branches containing only Seamless nodes should be considered as residual too !!!
void insertResidualScalingNodes(std::shared_ptr<GraphView> graphView)
{
    // TODO: double check this ...

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (isMerging(node))
        {
            int nbParents = node->getParents().size();
            for (int i = 0; i < nbParents; i++)
            {
                std::shared_ptr<Node> parentNode = node->getParent(i);
                bool parentIsForking = (parentNode->getChildren().size() > 1);

                if (parentIsForking)
                {
                    // temporary verbose ...

                    Log::info(" ### found residual branch at index {}", i);
                    Log::info(" ### inserting multiplicative node ...");

                    std::string residualNodeName = makeUniqueName(parentNode->name() + "_Res", graphView);
                    
                    // XXX
                    std::shared_ptr<Node> residualNode = Quantizer(1.0, residualNodeName);
                    addAttr(residualNode, "isActivationQuantizer");

                    residualNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                    residualNode->getOperator()->setBackend(determineBackend(parentNode));

                    graphView->insertParent(node, residualNode, i, 0, 0);
                }
            }
        }
    }
}

void insertScalingNodes(std::shared_ptr<GraphView> graphView)
{
    insertProducerScalingNodes(graphView);
    insertResidualScalingNodes(graphView);

    std::set<std::shared_ptr<Node>> nodeSet = graphView->getNodes();

    for (std::shared_ptr<Node> parentNode : nodeSet)
    {
        // Insert a Scaling node after each node that have to be quantized

        if (isAffine(parentNode) || isMerging(parentNode) || isNotQuantized(parentNode))
        {
            std::string scalingNodeName = makeUniqueName(parentNode->name() + "_Scaling", graphView);

            // XXX XXX XXX
            std::shared_ptr<Node> scalingNode = Quantizer(1.0, scalingNodeName);
            addAttr(scalingNode, "isActivationQuantizer");

            scalingNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
            scalingNode->getOperator()->setBackend(determineBackend(parentNode));

            if (parentNode->getChildren().size() > 0) {
                insertChildren(parentNode, scalingNode, graphView);
            } else {
                parentNode->addChild(scalingNode, 0, 0);
                graphView->add(scalingNode);
            }
            
            // In the case the node is a non-linear operator we want to add an extra
            // scaling node before it to rescale it's input ...

            if (isNotQuantized(parentNode))
            {
                std::string prevScalingNodeName = makeUniqueName(parentNode->name() + "_PrevScaling", graphView);

                // XXX XXX XXX
                std::shared_ptr<Node> prevScalingNode = Quantizer(1.0, prevScalingNodeName);
                addAttr(prevScalingNode, "isActivationQuantizer");

                prevScalingNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                prevScalingNode->getOperator()->setBackend(determineBackend(parentNode));

                graphView->insertParent(parentNode, prevScalingNode, 0, 0, 0);
            }
        }
    }
}

void normalizeParameters(std::shared_ptr<GraphView> graphView)
{
    // CREATE THE ACCUMULATED RATIO MAP ///////////////////////////////////////

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    std::unordered_map<std::shared_ptr<Node>, double> accumulatedRatios;
    for (std::shared_ptr<Node> node : nodeVector)
        accumulatedRatios.insert(std::make_pair(node, 1.0));

    // ITERATE OVER THE GRAPH /////////////////////////////////////////////////

    std::shared_ptr<Node> firstNode = getFirstNode(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        // Scaling nodes still have a ratio of 1, so they are seamless ...
        if (node->type() == "ReLU" || hasAttr(node, "isActivationQuantizer") || isSeamless(node))
        {
            if (node != firstNode)
            {
                std::shared_ptr<Node> prevNode = node->getParent(0);
                accumulatedRatios[node] = accumulatedRatios[prevNode];
            }
        }

        // Residual nodes should enter in this category but their ratio is 1 ...
        if (isAffine(node))
        {
            // Rescale the weight tensor
            
            std::shared_ptr<Tensor> weightTensor = getWeightTensor(node);

            double ratio = 1.0 / getTensorAbsoluteMax(weightTensor);

            // rescaleTensor(weightTensor, ratio);
            multiplyScalingFactor(node->getParent(1), ratio);

            // Accumulate the ratio

            if (node == firstNode) {
                accumulatedRatios[node] = ratio;
            } else {
                std::shared_ptr<Node> prevNode = node->getParent(0);
                accumulatedRatios[node] = accumulatedRatios[prevNode] * ratio;
            }

            // Handle the bias ..

            if (nodeHasBias(node))
            {
                std::shared_ptr<Tensor> biasTensor = getBiasTensor(node);
                //rescaleTensor(biasTensor, accumulatedRatios[node] );
                multiplyScalingFactor(node->getParent(2), accumulatedRatios[node]);
            }
        }

        if (isNotQuantized(node))
        {
            // Gather the previous scaling factor

            std::shared_ptr<Node> prevScalingNode = getPreviousScalingNode(node);
            double prevRatio = accumulatedRatios[prevScalingNode];

            // Cancel the accumulated ratio

            multiplyScalingFactor(prevScalingNode, 1 / prevRatio);

            // Revert the canceling by using the next scaling node

            accumulatedRatios[node] = prevRatio;
            std::shared_ptr<Node> nextScalingNode = getUniqueChild(node);
            multiplyScalingFactor(nextScalingNode, prevRatio);
        }

        if (isMerging(node))
        {
            if (node->type() == "MatMul")
            {
                // Multiply the input scaling factors !

                double leftRatio  = accumulatedRatios[node->getParent(0)];
                double rightRatio = accumulatedRatios[node->getParent(1)];

                accumulatedRatios[node] = leftRatio * rightRatio;
            }
            else
            {
                // Use a maximum arbitration !

                std::vector<std::shared_ptr<Node>> mergingNodes = node->getParents();

                // Compute the max ratio ...

                double maxRatio = 0;
                for (std::shared_ptr<Node> mergingNode : mergingNodes)
                {
                    double merginNodeRatio = accumulatedRatios[mergingNode];
                    if (merginNodeRatio > maxRatio)
                        maxRatio = merginNodeRatio;
                }

                accumulatedRatios[node] = maxRatio;

                // Rescale the previous scaling Nodes
                for (std::shared_ptr<Node> mergingNode : mergingNodes)
                {
                    double mergingNodeRatio = accumulatedRatios[mergingNode];
                    double rescaling = mergingNodeRatio / maxRatio;

                    std::shared_ptr<Node> scalingNode = getPreviousScalingNode(mergingNode);

                    multiplyScalingFactor(scalingNode, 1 / rescaling);

                    accumulatedRatios[mergingNode] /= rescaling; // optional ...
                }
            }
        }
    }
}

std::unordered_map<std::shared_ptr<Node>, double> computeRanges(std::shared_ptr<GraphView> graphView, std::vector<std::shared_ptr<Tensor>> calibrationSet, bool scalingNodesOnly, bool useCuda)
{
    std::unordered_map<std::shared_ptr<Node>, double> valueRanges;
    std::set<std::shared_ptr<Node>> nodeSet = graphView->getNodes();
    
    for (std::shared_ptr<Node> node : nodeSet)
        if ((scalingNodesOnly && hasAttr(node, "isActivationQuantizer")) || (!scalingNodesOnly && (node->type() != "Producer")))
            valueRanges.insert(std::make_pair(node, 0));

    if (useCuda)
        graphView->setBackend("cuda");

    SequentialScheduler scheduler(graphView);
    scheduler.resetScheduling();

    int it = 0;

    for (std::shared_ptr<Tensor> sample : calibrationSet)
    {
        //Log::info(" IT : {}", it++);

        // Inference ...

        if (useCuda)
            sample->setBackend("cuda");

        scheduler.forward(true, {sample});

        // Gather the sample ranges ...

        std::unordered_map<std::shared_ptr<Node>, double> sampleRanges;
        for (std::shared_ptr<Node> node : nodeSet)
        {
            if ((scalingNodesOnly && hasAttr(node, "isActivationQuantizer")) || (!scalingNodesOnly && (node->type() != "Producer")))
            {
                std::shared_ptr<Operator> nodeOperator = node->getOperator();
                std::shared_ptr<Tensor> valueTensor = std::static_pointer_cast<Tensor> (nodeOperator->getRawOutput(0));
                
                if (useCuda)
                    valueTensor->setBackend("cpu");

                double range = getTensorAbsoluteMax(valueTensor);

                // Associate the value to the scaling node ...
                sampleRanges.insert(std::make_pair(node, range));

                if (useCuda)
                    valueTensor->setBackend("cuda");
            }
        }

        // Update the global value ranges ...

        for (std::shared_ptr<Node> node : nodeSet)
        {
            if ((scalingNodesOnly && hasAttr(node, "isActivationQuantizer")) || (!scalingNodesOnly && (node->type() != "Producer")))
                if (sampleRanges[node] > valueRanges[node])
                    valueRanges[node] = sampleRanges[node];
        }

        if (useCuda)
            sample->setBackend("cpu");
    }   

    if (useCuda)
        graphView->setBackend("cpu");

    return valueRanges;
}

void normalizeActivations(std::shared_ptr<GraphView> graphView, std::unordered_map<std::shared_ptr<Node>, double> valueRanges)
{
    std::shared_ptr<Node> firstNode = getFirstNode(graphView);

    // CREATE THE ACCUMULATED RATIO MAP ///////////////////////////////////////

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    std::unordered_map<std::shared_ptr<Node>, double> accumulatedRatios;

    for (std::shared_ptr<Node> node : nodeVector)
        accumulatedRatios.insert(std::make_pair(node, 1.0));

    // ITERATE OVER THE GRAPH /////////////////////////////////////////////////

    for (std::shared_ptr<Node> node : nodeVector)
    {
        // Seamless scaling factor propagation ...

        if (isAffine(node) || isSeamless(node) || node->type() == "ReLU") 
        {
            if (node == firstNode) {
                accumulatedRatios[node] = 1.0;
            } else {
                std::shared_ptr<Node> prevNode = node->getParent(0);
                accumulatedRatios[node] = accumulatedRatios[prevNode];
            }
        }

        // Use the Scaling nodes to rescale the ranges ...

        if (hasAttr(node, "isActivationQuantizer")) 
        {
            std::shared_ptr<Node> prevNode = node->getParent(0);

            double prevRatio = accumulatedRatios[prevNode];
            double nodeRange = valueRanges[node]; 

            multiplyScalingFactor(node, prevRatio / nodeRange);

            accumulatedRatios[node] = nodeRange;

            // If prevNode is Affine, fix the bias ...

            if (isAffine(prevNode))
                if (nodeHasBias(prevNode))  
                    multiplyScalingFactor(prevNode->getParent(2), 1.0 / prevRatio);
        }

        // Merging nodes handling : use a maximum arbritration ...
        
        if (isMerging(node))
        {
            if (node->type() == "MatMul")
            {
                double leftRatio  = accumulatedRatios[node->getParent(0)];
                double rightRatio = accumulatedRatios[node->getParent(1)];
                accumulatedRatios[node] = leftRatio * rightRatio;
            }
            else
            {
                std::vector<std::shared_ptr<Node>> mergingNodes = node->getParents();

                // Compute the max ratio ...

                double maxRatio = 0;
                for (std::shared_ptr<Node> mergingNode : mergingNodes)
                {
                    double mergingNodeRatio = accumulatedRatios[mergingNode];
                    if (mergingNodeRatio > maxRatio)
                        maxRatio = mergingNodeRatio;
                }

                accumulatedRatios[node] = maxRatio;

                for (std::shared_ptr<Node> mergingNode : mergingNodes)
                {
                    double mergingNodeRatio = accumulatedRatios[mergingNode];
                    std::shared_ptr<Node> scalingNode = getPreviousScalingNode(mergingNode);
                    multiplyScalingFactor(scalingNode, mergingNodeRatio / maxRatio);
                    // Log::notice(" SCALING NODE : {} {}", scalingNode->type(), scalingNode->name());
                }
            }
        }

        if (isNotQuantized(node))
        {
            std::shared_ptr<Node> prevScalingNode = node->getParent(0);
            double prevRatio = accumulatedRatios[prevScalingNode];
            Log::notice(" prev ratio : {} ", prevRatio);

            // This causes the previous range to not full fill the [-1, 1] interval !!!
            // It could be avoided by systematicly add an extra Scaling node before each 
            // non linearity ...

            multiplyScalingFactor(prevScalingNode, prevRatio);
        }
    }
}

std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> computeSignMap(std::shared_ptr<GraphView> graphView, bool verbose)
{
    std::shared_ptr<Node> firstNode = getFirstNode(graphView);

    std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> signMap;

    std::pair<bool, bool> unsignedPair(true, true);
    for (std::shared_ptr<Node> node : graphView->getNodes())
        if (node->type() != "Producer") // XXX XXX XXX we should use nodeVector instead ...
            signMap.insert(std::make_pair(node, unsignedPair));

    // ITERATE OVER THE GRAPH

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        bool isFirstNode = (node == firstNode);

        if (isAffine(node))
        {   
            // Affine nodes always have a single parent 
            if (!isFirstNode) 
                signMap[node].first = signMap[node->getParent(0)].second;
            else
                signMap[node].first = false;

            signMap[node].second = false;
        } 

        if (hasAttr(node, "isActivationQuantizer")) 
        {
            signMap[node].second = false;

            // Scaling nodes always have a single parent 
            std::shared_ptr<Node> parent = node->getParent(0); 

            bool allChildrenAreReLU = true;
            allChildrenAreReLU &= !(node->getChildren().empty()); // a bit convoluted ...
            for (std::shared_ptr <Node> child : node->getChildren())
                allChildrenAreReLU &= (child->type() == "ReLU");

            // Correct the previous single node (when it is an Affine node) ...
            if (allChildrenAreReLU)
                if (isAffine(parent) || isMerging(parent))
                    signMap[parent].second = true;

            // Maintain unsigned output
            if (signMap[parent].second)
                signMap[node].second = true;

            // Set the link ...
            signMap[node].first = signMap[parent].second;
        }
       
        if (isMerging(node))
        {
            std::vector<std::shared_ptr<Node>> parentNodes = node->getParents(); 

            bool allParentAreSigned = true;
            bool allParentAreUnsigned = true;
            for(std::shared_ptr<Node> parent : parentNodes)
            {
                bool parentSign = signMap[parent].second;
                allParentAreSigned &= !parentSign;
                allParentAreUnsigned &= parentSign;
            }

            if (allParentAreSigned)
                signMap[node] = std::make_pair(false, false);
            else if (allParentAreUnsigned)
                signMap[node] = std::make_pair(true, true);
            else 
            {
                // Arbitration : Signed type wins !
                for(std::shared_ptr<Node> parent : parentNodes)
                {
                    while (!hasAttr(parent, "isActivationQuantizer"))
                    {
                        signMap[parent] = std::make_pair(false, false);
                        // We are on a branch so nodes always have 1 parent ...
                        parent = parent->getParent(0); 
                    }

                    signMap[parent].second = false;
                }

                signMap[node].first = false;
            }
        }
        
        if (node->type() == "ReLU" || isSeamless(node) || isNotQuantized(node))
        {
            // Thoses nodes always have a single parent 
            std::shared_ptr<Node> parent = node->getParent(0); 
            if (parent)
            {
                signMap[node].first = signMap[parent].second;
                signMap[node].second = signMap[node].first;
            }
            
        }
    }

    // VERBOSE

    if (verbose)
    {
        Log::info(" === SIGN MAP === ");
        for (std::shared_ptr<Node> node : nodeVector)
            Log::info(" {}{} | {}", static_cast<int>(signMap[node].first), static_cast<int>(signMap[node].second), node->name());
    }

    // SANITY CHECK (TEMPORARY)

    for (std::shared_ptr<Node> node : nodeVector)
    {
        for (std::shared_ptr<Node> child : node->getChildren()) 
        {
            if (signMap[node].second != signMap[child].first)
                Log::error(" computeSignMap : link is not sane ! ({} -> {})", node->name(), child->name());                    
        }
    }
    
    return signMap;
}

void quantizeNormalizedNetwork(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, bool noQuant, bool optimizeSigns, bool verbose)
{
    if (optimizeSigns && noQuant)
        AIDGE_THROW_OR_ABORT(std::runtime_error, " Sign-optimization can not be applied if network is not fully quantized ...");

    double signedMax = (1 << (nbBits - 1)) - 1;
    double unsignedMax = (1 << nbBits) - 1;

    std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> signMap;

    if (optimizeSigns)
        signMap = computeSignMap(graphView, verbose);
    else
    {
        // XXX XXX XXX we should use the (retreive) node vector
        std::pair<bool, bool> signedPair(false, false);
        for (std::shared_ptr<Node> node : graphView->getNodes())
            if (node->type() != "Producer")
                signMap.insert(std::make_pair(node, signedPair));
    }

    // ITERATE OVER THE GRAPH /////////////////////////////////////////////////

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (isAffine(node))
        {
            // Rescale the weight tensor
            multiplyScalingFactor(node->getParent(1), signedMax);

            // UUU Quantize the Producer !!!
            if (!noQuant)
                appendRoundClip(node->getParent(1), -(signedMax + 1), signedMax);

            // Rescale the bias tensor
            if (nodeHasBias(node))  
            {
                bool inputIsUnsigned = signMap[node].first;
                double rescaling = inputIsUnsigned ? unsignedMax * signedMax : signedMax * signedMax;
            
                multiplyScalingFactor(node->getParent(2), rescaling);

                // XXX TODO : enhance this ! 
                int biasMax = (1 << (12 + nbBits));
                if (!noQuant)
                    appendRoundClip(node->getParent(2), -(biasMax + 1), biasMax);
            }

            // Compensate the rescaling using the next Scaling node

            double rescaling = 1.0 / signedMax;

            bool inputIsUnsigned  = signMap[node].first; 
            bool outputIsUnsigned = signMap[node].second;

            rescaling /= inputIsUnsigned  ? unsignedMax : signedMax;
            rescaling *= outputIsUnsigned ? unsignedMax : signedMax;
            
            std::shared_ptr<Node> scalingNode = getUniqueChild(node); // TODO : assert if scalingNode is a Scaling ...

            multiplyScalingFactor(scalingNode, rescaling);          
        }
        
        if (isMerging(node))
        {
            double rescaling = 1.0;

            bool inputIsUnsigned  = signMap[node].first;
            bool outputIsUnsigned = signMap[node].second;

            rescaling /= inputIsUnsigned  ? unsignedMax : signedMax;
            rescaling *= outputIsUnsigned ? unsignedMax : signedMax;

            std::shared_ptr<Node> scalingNode = getUniqueChild(node); // TODO : assert if scalingNode is a Scaling ...
        
            // TODO : double check this ...
            if (node->type() == "MatMul")
                rescaling /= inputIsUnsigned ? unsignedMax : signedMax;

            multiplyScalingFactor(scalingNode, rescaling);          
        }

        if (isNotQuantized(node))
        {
            double rescaling = 1 / signedMax;  // XXX handle the signs !!!

            std::shared_ptr<Node> prevScalingNode = node->getParent(0);
            multiplyScalingFactor(prevScalingNode, rescaling);

            std::shared_ptr<Node> nextScalingNode = getUniqueChild(node);
            multiplyScalingFactor(nextScalingNode, 1 / rescaling);
        }
        
        // Handle the Scaling Nodes ...

        if (hasAttr(node, "isActivationQuantizer")) 
        {
            // Don't touch the scalings that precede non-linearities ...

            bool precedesNonLinearNode = false;
            if (node->getChildren().size() == 1)
                if (isNotQuantized(getUniqueChild(node)))
                    precedesNonLinearNode = true; 

            if (!noQuant && !precedesNonLinearNode) 
            {  
                // we need to gather the sign informations before we modify 
                // the node pointer with appendRoundClip() ... 

                bool inputIsUnsigned  = signMap[node].first;
                bool outputIsUnsigned = signMap[node].second;

                appendRoundClip(node, -(signedMax + 1), signedMax);

                if (optimizeSigns)
                {
                    double rescaling = 1.0;

                    rescaling /= inputIsUnsigned  ? unsignedMax : signedMax;
                    rescaling *= outputIsUnsigned ? unsignedMax : signedMax;

                    // XXX XXX XXX
                    multiplyScalingFactor(node, rescaling);

                    if (outputIsUnsigned)
                        setClipRange(node, 0, unsignedMax);                                  
                }
            }
        }
    }
}

void insertCompensationNodes(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, bool noQuant)
{
    // XXX Use the signMap to increase the resolution when possible ...
    double signedMax = (1 << (nbBits - 1)) - 1;    

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        // The appropriate strategy is to check if the Quantizer is not   
        // preceded by an Weighted node (that is not forking), and insert  
        // a mul node (Compensation) before it if so ...
 
        if (node->type() == "Quantizer")
        {
            // Note : this works because a Quantizer has only one Parent ...

            std::shared_ptr<Node> parentNode = node->getParent(0);
            bool parentHasWeight = isAffine(parentNode);
            bool parentIsForking = (parentNode->getChildren().size() > 1);

            if (parentIsForking || !parentHasWeight) // insert a Compensation Node ...
            {   
                // Create and insert the multiplicative node before the Quantizer

                std::string mulNodeName = makeUniqueName(node->name() + "_Mul", graphView);
                std::shared_ptr<Node> mulNode = Mul(mulNodeName);
                
                // XXX XXX XXX addAttr(mulNode, "isCompensation");

                mulNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                mulNode->getOperator()->setBackend(determineBackend(node));

                graphView->insertParent(node, mulNode, 0, 0, 0);

                // Add the coeff producer to the multiplier node

                std::shared_ptr<Node> coeffProducer = addProducer(mulNode, 1, {1}, ""); 
                std::shared_ptr<Tensor> coeffTensor = std::make_shared<Tensor>(Array1D<double, 1> {signedMax});
                coeffProducer->getOperator()->setOutput(0, coeffTensor);

                coeffProducer->getOperator()->setDataType(DataType::Float64);
                coeffProducer->getOperator()->setBackend(determineBackend(node)); 

                graphView->add(coeffProducer); // needed ?

                // Adapt the scaling factor value accordingly

                multiplyScalingFactor(node, 1.0 / signedMax); // XXX XXX XXX OK

                // Insert a Quantizer for the coeffProducer that will handle  
                // the single-shift approximation via it's scalingFactor ...

                insertScalingBelowProducer(coeffProducer, graphView);
                
                if (!noQuant) 
                {
                    // XXX XXX XXX double check this ...
                    std::shared_ptr<Node> coeffQuantizer = mulNode->getParent(1);
                    appendRoundClip(coeffQuantizer, -(signedMax + 1), signedMax);
                }

            }
        }
    }
}

void performSingleShiftApproximation(std::shared_ptr<GraphView> graphView)
{
    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (node->type() == "Quantizer")
        {
            std::shared_ptr<Node> linearNode = node->getParent(0);

            double base = getScalingFactor(node);
            double approx = std::pow(2, std::ceil(std::log2(base)));
            double ratio = approx / base;

            // set the scaling factor value to the approximation ...

            multiplyScalingFactor(node, ratio);

            // compensate the ratio using the previous node scaling factors ...

            multiplyScalingFactor(linearNode->getParent(1), 1.0 / ratio);
            if (nodeHasBias(linearNode))
                multiplyScalingFactor(linearNode->getParent(2), 1.0 / ratio);
        }
    }
}

static void printScalingFactors(std::shared_ptr<GraphView> graphView)
{
    for (auto node : retrieveNodeVector(graphView))
        if (hasAttr(node, "isActivationQuantizer") || node->type() == "Quantizer")
        {
            double scalingFactor = getScalingFactor(node);
            Log::notice(" SCALING FACTOR : {} ({})", scalingFactor, node->name());
        }
}

static void setupDataType(std::shared_ptr<GraphView> graphView, std::vector<std::shared_ptr<Tensor>> calibrationSet, DataType dataType)
{
    graphView->setDataType(dataType);

    for (auto inputNode : graphView->inputNodes()) {
        auto op = std::static_pointer_cast<OperatorTensor>(inputNode->getOperator());
        auto inputTensor = op->getInput(0);
        if (inputTensor)
            inputTensor->setDataType(dataType);
    }

    for (auto tensor : calibrationSet)
        tensor->setDataType(dataType);
}

void quantizeNetwork(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, std::vector<std::shared_ptr<Tensor>> calibrationSet, DataType targetType, Clipping clippingMode, bool noQuant, bool optimizeSigns, bool singleShift, bool useCuda, bool foldGraph, bool verbose)
{
    Log::notice(" === QUANT PTQ 0.3.0 === ");

    graphView->setBackend("cpu");

    if (!checkArchitecture(graphView))
        return;

    DataType initialDataType = (calibrationSet[0])->dataType();
    setupDataType(graphView, calibrationSet, DataType::Float64);

    Log::notice(" Preparing the network for the PTQ ... ");
    prepareNetwork(graphView);

    Log::notice(" Inserting the scaling nodes ...");
    insertScalingNodes(graphView);

    // TODO : double check the CLE ...
    crossLayerEqualization(graphView); // XXX XXX XXX

    Log::notice(" Normalizing the parameters ...");
    normalizeParameters(graphView);

    Log::notice(" Computing the value ranges ...");
    std::unordered_map<std::shared_ptr<Node>, double> valueRanges = computeRanges(graphView, calibrationSet, true, useCuda);

    Log::notice(" Optimizing the clipping values ...");
    valueRanges = adjustRanges(clippingMode, valueRanges, nbBits, graphView, calibrationSet, useCuda, verbose);

    Log::notice(" Normalizing the activations ...");
    normalizeActivations(graphView, valueRanges);

    Log::notice(" Quantizing the normalized network ...");
    quantizeNormalizedNetwork(graphView, nbBits, noQuant, optimizeSigns, verbose);

    if (singleShift)
    {
        Log::notice( " Inserting the compensation nodes ...");
        insertCompensationNodes(graphView, nbBits, noQuant);

        Log::notice(" Performing the Single-Shift approximation ...");
        performSingleShiftApproximation(graphView);
    }

    Log::notice(" Casting the network to the target type ({}) ...", targetType);
    castQuantizedNetwork(graphView, targetType, singleShift);

    if (foldGraph)
    {
        Log::notice(" Folding the Producer's Quantizers ...");
        foldProducerQuantizers(graphView);
    }

    // TODO ...
    // Log::notice(" Clearing the input nodes ...");

    if (verbose)
        printScalingFactors(graphView);

    if (useCuda)
        graphView->setBackend("cuda");

    Log::notice(" Reseting the scheduler ...");
    SequentialScheduler scheduler(graphView);
    scheduler.resetScheduling();

    Log::notice(" Network is quantized !");
}

std::unordered_map<std::string, double> getWeightRanges(std::shared_ptr<GraphView> graphView)
{
    std::unordered_map<std::string, double> weightRanges;

    for (std::shared_ptr<Node> node : graphView->getNodes())
    {
        if (isAffine(node))
        {
            std::shared_ptr<Tensor> weightTensor = getWeightTensor(node);
            double range = getTensorAbsoluteMax(weightTensor);
            weightRanges.insert(std::make_pair(node->name(), range));
        }
    }

    return weightRanges;
}

void clearBiases(std::shared_ptr<GraphView> graphView) 
{
    for (std::shared_ptr<Node> node : graphView->getNodes()) {
        if (node->type() == "FC" || node->type() == "Conv2D") {
            std::shared_ptr<Tensor> biasTensor = std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(2);
            //rescaleTensor(biasTensor, 0);
            //insertScalingBelowProducer(node->getParent(2), 0, graphView);
            multiplyScalingFactor(node->getParent(2), 0);
        }
    }
}
}
