/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/quantization/QAT/QAT_LSQ.hpp"
#include "aidge/operator/LSQ.hpp"
#include "aidge/operator/ReLU.hpp"


#include "aidge/data/Tensor.hpp"
#include "aidge/graph/GraphView.hpp"
#include "aidge/scheduler/SequentialScheduler.hpp"
#include "aidge/scheduler/Scheduler.hpp"
#include "aidge/graph/Matching.hpp"
#include "aidge/recipes/QuantRecipes.hpp"


namespace Aidge 
{

static float getTensorAbsMean(std::shared_ptr<Tensor> tensor)
{
    auto valueTensor = (*tensor).abs().mean();
    std::shared_ptr<Tensor> fallback;
    const Tensor& localTensor = valueTensor.refCastFrom(fallback, DataType::Float32, "cpu");
    return localTensor.get<float>(0);
}

static float getTensorStd(std::shared_ptr<Tensor> tensor)
{
    auto valueTensor = (*tensor);
    
    auto skewedTensor = valueTensor - valueTensor.mean();
    auto squaredTensor = skewedTensor * skewedTensor;
    auto varianceTensor = squaredTensor.mean();

    std::shared_ptr<Tensor> fallback;
    auto localTensor = varianceTensor.refCastFrom(fallback, DataType::Float32, "cpu");
    
    float variance = localTensor.get<float>(0);
    return std::sqrt(variance);
}


// INIT THE STEP SIZE OF A QUANTIZER NODE

static bool initStepSize(std::shared_ptr<Node> quantizer)
{
    const auto quantizerOp = std::static_pointer_cast<LSQ_Op>(quantizer->getOperator());

    // This formula is the one proposed in the paper ...

    // float inputAbsMean = getTensorAbsMean(quantizerOp->getInput(0));
    // float stepSize = 2.0f * (inputAbsMean / std::sqrt(quantizerOp->range().second));

    // .. but this formula seems to work better !!!

    float inputStd = getTensorStd(quantizerOp->getInput(0));
    float stepSize = 8.0f * (inputStd / (quantizerOp->range().second));

    // TODO : use the scalar constructor
    auto stepSizeTensor = std::make_shared<Tensor>(Array1D<float, 1>({{stepSize}})); 

    // XXX Manage backend here ?
    stepSizeTensor->setBackend(quantizerOp->getInput(0)->backend());
    stepSizeTensor->setDataType(quantizerOp->getInput(0)->dataType());

    auto stepSizeProducer = quantizer->getParent(1);

    stepSizeProducer->getOperator()->setOutput(0, stepSizeTensor);

    Log::notice(" [ INIT STEP SIZE = {} ] ", stepSize);

    return false;
}

static void setupInputQuantizers(std::shared_ptr<GraphView> graphView, size_t nbBits)
{
    const auto matches = SinglePassGraphMatching(graphView).match("(Conv2D#|PaddedConv2D#|FC#)");

    for (const auto& match : matches) 
    {
        auto linearNode = match.graph->rootNode(); 

        // Log::notice(" SET INPUT QUANTIZER : {} ", linearNode->type());

        std::pair<int, int> signedRange = {-std::pow(2, nbBits - 1), std::pow(2, nbBits - 1) - 1};
        std::pair<int, int> unsignedRange = {0, std::pow(2, nbBits) - 1};

        // Create the input quantizer node

        auto quantizerName = makeUniqueName(linearNode->name() + "_lsq_i", graphView);  
        auto quantizerNode = LSQ(signedRange, quantizerName);

        // Init the step-size using the node call stack

        quantizerNode->addBeforeForward([quantizerNode](){ return initStepSize(quantizerNode); });

        // Absorb the ReLU when possible ...

        bool nodeHasParent = static_cast<bool> (linearNode->getParents()[0]);  // XXX is this safe ?

        if (nodeHasParent) 
        {
            bool allParentsAreReLU = true;
            for (auto parentNode : linearNode->getParents())
                if (parentNode->type() != "ReLU")
                    allParentsAreReLU = false;

            if (allParentsAreReLU) {
                auto quantizerOp = std::static_pointer_cast<LSQ_Op> (quantizerNode->getOperator());
                quantizerOp->range() = unsignedRange;
            }

            // TODO : remove the ReLUs when possible
        }

        // Insert the quantizer in the graphView ...
        // (We need to handle the case where the linear node is the first one)

        if (nodeHasParent) {
            graphView->insertParent(linearNode, quantizerNode, 0, 0, 0);
        } else {
            quantizerNode->addChild(graphView);
            graphView->add(quantizerNode);
        }
    }
}

// PARAM QUANTIZERS INSERTION

static void setupParamQuantizers(std::shared_ptr<GraphView> graphView, size_t nbBits)
{
    const auto matches = SinglePassGraphMatching(graphView).match("(Conv2D#|PaddedConv2D#|FC#)");

    std::pair<int, int> signedRange = {-std::pow(2, nbBits - 1), std::pow(2, nbBits - 1) - 1};

    for (const auto& match : matches) 
    {       
        auto linearNode = match.graph->rootNode(); 

        // Log::notice(" SET PARAM QUANTIZER : {} ", linearNode->type());

        // TODO : double check this, and use createUniqueName()
        auto quantizerName = makeUniqueName(linearNode->name() + "_lsq_p", graphView);  
        auto quantizerNode = LSQ(signedRange, quantizerName); 

        // Init the step-size using the node call stack

        quantizerNode->addBeforeForward([quantizerNode](){ return initStepSize(quantizerNode); });

        // Insert the quantizer in the graphView

        graphView->insertParent(linearNode, quantizerNode, 1, 0, 0);
    }
}

void QuantLSQ::setupQuantizers(std::shared_ptr<GraphView> graphView, size_t nbBits)
{
    sanitizeNodeNames(graphView);
    setupInputQuantizers(graphView, nbBits);
    setupParamQuantizers(graphView, nbBits);
}

}