/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/operator/PTQMetaOps.hpp"

#include <memory>
#include <string>
#include <utility>

#include "aidge/operator/Clip.hpp"
#include "aidge/operator/Mul.hpp"
#include "aidge/operator/Round.hpp"
#include "aidge/operator/BitShift.hpp"
#include "aidge/operator/Cast.hpp"

#include "aidge/graph/Node.hpp"
#include "aidge/graph/OpArgs.hpp"
#include "aidge/operator/MetaOperator.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/ArrayHelpers.hpp"
#include "aidge/utils/Types.h"
#include "aidge/operator/Identity.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/Log.hpp"

namespace Aidge
{ 

static bool hasAttr(std::shared_ptr<Aidge::Node> node, std::string attr)
{
    return node->attributes()->hasAttr("quantization.ptq." + attr);
}

static void addAttr(std::shared_ptr<Aidge::Node> node, std::string attr, double value = 0.0)
{
    node->attributes()->addAttr("quantization.ptq." + attr, value);
}

// TODO : rework this
static void copyDynamicAttributes(std::shared_ptr<Aidge::Node> prevNode, std::shared_ptr<Aidge::Node> newNode)
{
    if (hasAttr(prevNode, "isProducerQuantizer"))
        addAttr(newNode, "isProducerQuantizer");

    if (hasAttr(prevNode, "isActivationQuantizer"))
        addAttr(newNode, "isActivationQuantizer");  
}

std::shared_ptr<Node> Quantizer(double scalingFactor, const std::string& name)
{
    std::shared_ptr<Node> mulNode =  Mul(name + "_MulQuant");

    // Scaling Factor Producer

    std::shared_ptr<Tensor> scalingFactorTensor = std::make_shared<Tensor>(Array1D<double, 1> {scalingFactor});
    std::shared_ptr<Node> scalingFactorProducer = addProducer<1>(mulNode, 1, {1}, "ScalingFactor");
    scalingFactorProducer->getOperator()->setOutput(0, scalingFactorTensor);

    // TODO : the above could be replaced by :
    // std::shared_ptr<Node> scalingFactorProducer = Producer(scalingFactorTensor);
    // scalingFactorProducer->addChild(mulNode, 0, 1);

    // create the graphView ...

    std::shared_ptr<GraphView> graphView = Sequential({mulNode});
    graphView->add(scalingFactorProducer);

    // alternative : capture the Producer ...
    // std::shared_ptr<GraphView> connectedGraphView = getConnectedGraphView(mulNode); 

    std::shared_ptr<Node> quantizer = MetaOperator("Quantizer", graphView, {}, name); // an simpler prototype exists ...

    return quantizer;
}

void multiplyScalingFactor(std::shared_ptr<Aidge::Node> quantizer, double coeff)
{
    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());

    // Get the Mul node from the microGraph

    std::shared_ptr<Node> mulNode = nullptr;
    auto microGraph = quantizerOp->getMicroGraph();
    for (auto node : microGraph->getNodes())
        if (node->type() == "Mul")
            mulNode = node;

    // Retreive the previous scaling factor

    auto scalingFactorTensor = std::static_pointer_cast<OperatorTensor>(mulNode->getOperator())->getInput(1);

    std::shared_ptr<Tensor> fallback;
    const Tensor& localTensor = scalingFactorTensor->refCastFrom(fallback, DataType::Float64, "cpu");
    double prevScalingFactor = localTensor.get<double>(0);    

    // Create the new scaling factor tensor

    std::shared_ptr<Tensor> newScalingFactorTensor = std::make_shared<Tensor>(prevScalingFactor * coeff);
    newScalingFactorTensor->setBackend(scalingFactorTensor->backend());
    newScalingFactorTensor->setDataType(scalingFactorTensor->dataType());

    // Set the tensor of the producer

    auto producer = mulNode->getParent(1);
    producer->getOperator()->setOutput(0, newScalingFactorTensor);
}

void appendRoundClip(std::shared_ptr<Node>& quantizer, double clipMin, double clipMax)
{
    // Retreive a clone of the microGraph

    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    auto microGraph = quantizerOp->getMicroGraph()->clone();

    // Save the datatype / backend

    auto outputNode = *(microGraph->outputNodes().begin());
    auto outputOp = std::static_pointer_cast<OperatorTensor> (outputNode->getOperator());

    auto dataType = outputOp->getOutput(0)->dataType();
    auto backend = outputOp->getOutput(0)->backend();

    // append round

    auto roundNode = Round(quantizer->name() + "_RoundQuant");
    outputNode->addChild(roundNode, 0, 0);
    microGraph->add(roundNode);

    // append clip

    auto clipNode = Clip(quantizer->name() + "_ClipQuant");

    auto minTensor = std::make_shared<Tensor>(clipMin);
    auto minNode = Producer(minTensor);
    minNode->addChild(clipNode, 0, 1);
    microGraph->add(minNode);

    auto maxTensor = std::make_shared<Tensor>(clipMax);
    auto maxNode = Producer(maxTensor);   
    maxNode->addChild(clipNode, 0, 2);
    microGraph->add(maxNode);

    roundNode->addChild(clipNode, 0, 0);
    microGraph->add(clipNode);

    // set the datatype / backend

    microGraph->setDataType(dataType);
    microGraph->setBackend(backend);

    // create the new quantizer and replace the previous one

    std::shared_ptr<Node> newQuantizer = MetaOperator("Quantizer", microGraph, {}, quantizer->name());
    copyDynamicAttributes(quantizer, newQuantizer);
    GraphView::replace({quantizer}, {newQuantizer});

    // replace the old pointer with the new one (by reference)

    quantizer = newQuantizer;
}

double getScalingFactor(std::shared_ptr<Node> quantizer)
{
    // Retreive the previous microGraph

    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    auto microGraph = quantizerOp->getMicroGraph();

    // Get the Mul node from the microGraph

    std::shared_ptr<Node> mulNode = nullptr;
    for (auto node : microGraph->getNodes())
        if (node->type() == "Mul")
            mulNode = node;

    auto mulOp = std::static_pointer_cast<OperatorTensor> (mulNode->getOperator()); 

    // Retreive the scaling factor

    auto scalingFactorTensor = mulOp->getInput(1);

    std::shared_ptr<Tensor> fallback;
    const Tensor& localTensor = scalingFactorTensor->refCastFrom(fallback, DataType::Float64, "cpu");
    double scalingFactor = localTensor.get<double>(0);

    return scalingFactor;
}

void setClipRange(std::shared_ptr<Node> quantizer, double min, double max)
{
    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    auto microGraph = quantizerOp->getMicroGraph();

    std::shared_ptr<Node> clipNode = nullptr;
    for (auto node : microGraph->getNodes())
        if (node->type() == "Clip")
            clipNode = node;

    // TODO : assert that we've got not a nullptr ...

    auto clipOp = std::static_pointer_cast<Clip_Op> (clipNode->getOperator()); 

    // set the attributes

    clipOp->max() = max;
    clipOp->min() = min;

    // Retreive the previous min/max tensors 

    auto minTensor = std::static_pointer_cast<OperatorTensor>(clipNode->getOperator())->getInput(1);
    auto maxTensor = std::static_pointer_cast<OperatorTensor>(clipNode->getOperator())->getInput(2);

    // Create the new min/max tensors

    std::shared_ptr<Tensor> newMinTensor = std::make_shared<Tensor>(min);
    newMinTensor->setBackend(minTensor->backend());
    newMinTensor->setDataType(minTensor->dataType());

    std::shared_ptr<Tensor> newMaxTensor = std::make_shared<Tensor>(max);
    newMaxTensor->setBackend(maxTensor->backend());
    newMaxTensor->setDataType(maxTensor->dataType());

    // Set the tensors of the producer

    auto minProducer = clipNode->getParent(1);
    minProducer->getOperator()->setOutput(0, newMinTensor);

    auto maxProducer = clipNode->getParent(2);
    maxProducer->getOperator()->setOutput(0, newMaxTensor);
}

void removeRound(std::shared_ptr<Node>& quantizer)
{
    // Retreive a clone of the microGraph

    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    auto microGraph = quantizerOp->getMicroGraph()->clone();

    // retreive the rounding node

    std::shared_ptr<Node> roundNode = nullptr;
    for (auto node : microGraph->getNodes())
        if (node->type() == "Round")
            roundNode = node;

    if (roundNode == nullptr)
        return;
    
    // remove the Round node

    microGraph->replace({roundNode}, {});

    // Create the new quantizer and replace the previous one 

    std::shared_ptr<Node> newQuantizer = MetaOperator("Quantizer", microGraph, {}, quantizer->name());
    copyDynamicAttributes(quantizer, newQuantizer);
    GraphView::replace({quantizer}, {newQuantizer});

    // replace the old pointer with the new one (by reference)

    quantizer = newQuantizer;  
}

void replaceScalingWithBitShift(std::shared_ptr<Node>& quantizer)
{
    // Retreive a clone of the microGraph

    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    auto microGraph = quantizerOp->getMicroGraph()->clone();

    // retreive the multiplicative (scaling) node

    std::shared_ptr<Node> mulNode = nullptr;
    for (auto node : microGraph->getNodes())
        if (node->type() == "Mul")
            mulNode = node;

    auto mulOp = std::static_pointer_cast<OperatorTensor> (mulNode->getOperator());

    // Save the datatype / backend

    auto dataType = mulOp->getOutput(0)->dataType();
    auto backend = mulOp->getOutput(0)->backend();

    // compute the shift value 

    double scalingFactor = getScalingFactor(quantizer);
    int bitShiftAmount = -std::round(std::log2(scalingFactor));
    auto bitShiftDirection = BitShift_Op::BitShiftDirection::right;

    Log::notice(" SHIFT AMOUNT = {} ({})", bitShiftAmount, scalingFactor);

    if (bitShiftAmount < 0 )
    {
        bitShiftDirection = BitShift_Op::BitShiftDirection::left;
        bitShiftAmount = -bitShiftAmount;
    }

    bool bitShiftRounding = true; // XXX use an argument !!!

    // create the replacement bit-shift nodes

    auto bitShiftNode = BitShift(bitShiftDirection, bitShiftRounding, quantizer->name() + "_BitShiftQuant"); 
    auto bitShiftTensor = std::make_shared<Tensor>(Array1D<int, 1> {bitShiftAmount});

    auto bitShiftProducer = Producer(bitShiftTensor, "bitShiftAmount");
    bitShiftProducer->addChild(bitShiftNode, 0, 1);

    // edit the micrograph

    microGraph->replace({mulNode, mulNode->getParent(1)}, {bitShiftNode, bitShiftNode->getParent(1)});

    // set the datatype / backend

    microGraph->setDataType(dataType);
    microGraph->setBackend(backend);

    // create the new quantizer and replace the previous one

    std::shared_ptr<Node> newQuantizer = MetaOperator("Quantizer", microGraph, {}, quantizer->name());
    copyDynamicAttributes(quantizer, newQuantizer);  
    GraphView::replace({quantizer}, {newQuantizer});

    // replace the old pointer with the new one (by reference)

    quantizer = newQuantizer;  
}

void castQuantizerIOs(std::shared_ptr<Node>& quantizer, Aidge::DataType externalType)
{
    // Retreive a clone of the microGraph

    auto quantizerOp = std::static_pointer_cast<MetaOperator_Op> (quantizer->getOperator());
    auto microGraph = quantizerOp->getMicroGraph()->clone();

    // Edit the micrograph (insert Cast nodes at it's IOs)

    auto mulNode = *(microGraph->inputNodes().begin()); // TODO : assert that mulNode is a Mul !
    auto mulOp = std::static_pointer_cast<OperatorTensor> (mulNode->getOperator());

    auto internalType = mulOp->getOutput(0)->dataType();
    auto castInputNode = Cast(internalType, quantizer->name() + "_CastIn");  
    auto castOutputNode = Cast(externalType, quantizer->name() + "_CastOut");

    microGraph = Sequential({castInputNode, microGraph, castOutputNode});

    // Set the micrograph datatype 

    microGraph->setDataType(internalType);
    castOutputNode->getOperator()->setDataType(externalType);

    // Set the micrograph backend 

    auto backend = mulOp->getOutput(0)->backend(); 
    microGraph->setBackend(backend); 

    // Create the new quantizer and replace the old one

    std::shared_ptr<Node> newQuantizer = MetaOperator("Quantizer", microGraph, {}, quantizer->name());
    copyDynamicAttributes(quantizer, newQuantizer);  
    GraphView::replace({quantizer}, {newQuantizer});

    // replace the old pointer with the new one (by reference)

    quantizer = newQuantizer;  
}

}